C   circles - calculate identified circles statistics from CMB data (cosmic topology)
C
C   Copyright (C) 2004 Boud Roukema
C
C   This program is free software; you can redistribute it and/or modify
C   it under the terms of the GNU General Public License as published by
C   the Free Software Foundation; either version 2, or (at your option)
C   any later version.
C
C   This program is distributed in the hope that it will be useful,
C   but WITHOUT ANY WARRANTY; without even the implied warranty of
C   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C   GNU General Public License for more details.
C
C   You should have received a copy of the GNU General Public License
C   along with this program; if not, write to the Free Software Foundation,
C   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
C


      subroutine circ_f(indir,nindir)

C     07.12.2004 In the alpha distributions of the package "circles", 
C     that is, for versions less than 1.0.0, a lot of work still is likely
C     to be necessary in improving internal documentation (in program files)
C     and external documentation (such as the texinfo file). You are most
C     welcome to help and integrate your improvements with other people
C     working on this GNU GPL protected package.   --boud (B. Roukema)

c     ==== B. Roukema 27/10/98, IUCAA  boud at iucaa.ernet.in
c     (1) work out positions of COBE "pixels" in fundamental polyhedron.
c     (2) calculate statistics of pairs around circles
c     ====

      parameter(pi=3.1415926535898)
      parameter(pi_180=pi/180.0)

C     INPUTS:
      character*(*)  indir     ! directory with input files
      integer        nindir    ! number of characters in  indir

C     INTERNAL:
      character*256  indir2 ! safe string version of indir

      logical*4  cmb           ! do COBE data (T) or simulation (F)?

      real*4   fun_eq(3,3)      ! fundamental axes in B1950.0 equatorial coords
      real*4   efund(3,3) ! galactic coords
      parameter(maxfp=30)       ! upper bound to  [2 rSLS/Rfund(i)]
      real*4   Rfund(3)         ! lengths of axes in h^-1 Mpc
c      real*4   xlis(3)          ! temporary storage

c     common with mkmap
c     common /topocomm/ Rx,Ry,Rz,np,g1,g2,efund,einv
      common /topocomm/ Rfund,np,g1,g2,efund,einv
      integer*4  np             ! in sum over k values

      parameter(maxp=10)
      real*4   g1(-maxp:maxp,-maxp:maxp,-maxp:maxp) ! amplitudes
      real*4   g2(-maxp:maxp,-maxp:maxp,-maxp:maxp) ! amplitudes

      real*4     einv(3,3)
c     real*8   enorm8
c      real*8   efund8(3,3),einv_8(3,3), xmat3_8(3,3)
c      real*8   vec_8(3),vec2_8(3)
c      real*8   efund8(4,4),einv_8(4,4), xmat4_8(4,4)
c      real*8   vec_8(4),vec2_8(4)

      
      real*4    q_0,H_0,cc_0, zSLS, specn
c      parameter(maxcurv=4)

      include 'lib/npar.dec'
      
      real*4    q(maxcurv),cc(maxcurv)
      real*4    omm(maxcurv)
      real*4    rSLS            ! distance to SLS in h^-1 Mpc
      real*4    drSLS           ! thickness of SLS in h^-1 Mpc
      common /cosmcomm/ q_0,H_0,cc_0, zSLS, specn, rSLS !

c     ==== COBE stuff in rdCOBE.f ====
c      parameter(npix=6144,maxn=8096) !! WARNING declared several times
      include 'lib/npix.dec'

      real*4   temp(maxn),errtemp(maxn) !   Temperature and uncert.
c     real*4   eclong(maxn),eclat(maxn) ! ecliptic
      real*4   gallong(maxn),gallat(maxn) ! galactic
      real*4   eqlong(maxn),eqlat(maxn) ! ra dec in degrees, degrees

c      real*4   xg(maxn),yg(maxn),zg(maxn) ! galactic x,y,z for unit radius

c      parameter(nxint=180,nyint=90)
      include 'lib/nxint.dec'
      integer*4 inear(nxint,nyint) ! nearest pixel number
      real*4    dinear(nxint,nyint)
      real*4    dxint,dyint
c     -- Eq. Cartesian coords of table --
      real*4    x(nxint,nyint),y(nxint,nyint),z(nxint,nyint) 

      common /cobecomm/  temp,errtemp,gallong,gallat,eqlong,eqlat,
     _     dsm,gb0,dxint,dyint,inear,dinear,x,y,z

      logical*4 measur,noise    
c      real*4    ranerr(maxn)
      common /noisecomm/ measur,noise


c     --- function declarations ---
      real*4    pofk,delT,delTgal,cobeDMR,testfn, dTg_nodip,wmaplin
      external  pofk,delT,delTgal,cobeDMR,testfn, dTg_nodip,wmaplin
      complex*16    delTrad8
      external  delTrad8
      real*8    dofz            ! in cosmdist-0.2.1 packge
      external  dofz

c     real*8    red_d           ! local value: double precision redshift
c     ====

c     --- for output of positions in tipsy format ---
c      real*4    xT2(npix,3)     ! coords in T^2 frame
c      real*4    dummy(npix,4)   ! dummies for tipsy, but (1,4)= 1 = mass
c      character*256   flnm        
      character*3     ch3
      
c      real*4          header(128) 

c     ---- difference statistics ----
c     parameter(mxsig=30)
c      parameter(mxpar=1)

c      include 'lib/npar.dec'  ! above

c     -- ipar loops through each individual topology hypothesis --
      real*4    sigma(mxpar,maxcurv),drsigma(mxpar),drs2(mxpar)
      real*4    dmean(mxpar,maxcurv)
      real*4    sfreq(mxpar,maxcurv)
      real*4    corr(mxpar,maxcurv),cwght(mxpar,maxcurv)
      real*4    cdopp(mxpar,maxcurv)

c     --- temporary copy while looping through theta values (dodec case)
      parameter(mxtheta= 200)
      real*4    sig_th(mxtheta)
      real*4    dme_th(mxtheta)
      real*4    sfr_th(mxtheta)
      real*4    cor_th(mxtheta),cwg_th(mxtheta)
      real*4    cdo_th(mxtheta)

c     --- especially useful for Poincare dodecahedral hypothesis ---
      real*4    gl_th(mxtheta),gb_th(mxtheta)
      real*4    thdod_th(mxtheta),ang_th(mxtheta)

      real*4    glpar(mxpar) ! g long
      real*4    gbpar(mxpar) ! g lat
      real*4    thpar(mxpar) ! theta of dodecahedron rotation
      real*4    angpar(mxpar) ! ang radius of circle

      parameter(maxpair=20000)
      parameter(mxdlist=1,idlist=1)
      real*4    dlist(maxpair,mxdlist,maxcurv)
      integer*4 npair(mxpar,maxcurv)

      integer*4 npa_th(mxtheta) ! temporary copy

      real*4    sigbig(maxcurv),sfrbig(maxcurv)

      logical*4 test,rotate
      
c     ---- galactic cut ----
      logical*4 galcut
      real*4    bcut            ! galactic latitude cut
      real*4    factcutnoise    ! factor mult. Ssigma above which pix ignored
c     set  factcutnoise  to  10.0 or greater to have no effect

c     ---- magellanic stream (including LMC, SMC) optional cut ---
      logical*4 isMage          ! function
      logical*4 antiGC          ! function - in anti GCentre region
      logical*4 SGcut           ! function - arbitrary South hemi cut
      logical*4 inospot         ! function - in Ophiuchus,Orion complex/loop?
      logical*4 inos            ! local value- in Ophiuchus,Orion complex/loop?

c     ---- smoothing ? ----
      logical*4 smooth,new,dosmap,dos2map,dos2mean

      logical*4 cdoppwarn

c     --- 
c      parameter(maxrun=1000)
c      real*4    ranlis(2,maxrun),xlong(maxrun)

c     --- save circle data  -- only evaluate for icc=1=irun
c      parameter(iimax=50,jjmax=50, maxtheta=100)
c      parameter(iimax=180,jjmax=180, maxtheta=360) ! 03.04.2003
c      parameter(iimax=50,jjmax=50, maxtheta=720) ! 03.04.2003
c      parameter(iimax=10,jjmax=10, maxtheta=720) ! 18.12.2003

      include 'lib/ncopies.dec'

      integer*4  imax,jmax
      integer*4    nthc(-iimax:iimax,0:jjmax)
      real*4    glatc(2,-iimax:iimax,0:jjmax), ! 2 centres
     _     glongc(2,-iimax:iimax,0:jjmax), ! 2 centres
     _     wmodc(-iimax:iimax,0:jjmax), ! dtheta * rSLS
     _     rSLSc,
     _     dthMpc(-iimax:iimax,0:jjmax) ! dtheta * rSLS


      real*4    tt1c(maxtheta, -iimax:iimax,0:jjmax)
      real*4    tt2c(maxtheta, -iimax:iimax,0:jjmax)
      real*4    s1c(maxtheta, -iimax:iimax,0:jjmax)
      real*4    s2c(maxtheta, -iimax:iimax,0:jjmax)
c     ---- 24.02.2004 ----
c     -- store galactic long, lat at each pixel for double-checking --
      real*4    glpix1(maxtheta, -iimax:iimax,0:jjmax)
      real*4    gbpix1(maxtheta, -iimax:iimax,0:jjmax)
      real*4    glpix2(maxtheta, -iimax:iimax,0:jjmax)
      real*4    gbpix2(maxtheta, -iimax:iimax,0:jjmax)
      


c     ---- alternative to the RE97 T2 hypothesis ----
      logical*4 doRE97 ! .true. if doing RE97
      logical*4 doRfund ! .true. if modifying Rfund3 scale
      logical*4 dopolerot ! .true. if rotating about the long axis
      logical*4 donudge ! .true. if nudging around a position by few deg.
      logical*4 donud3 ! .true. if 2nd nudge param is th_d0
      real*4    thnud
c      logical*4 doranlong ! .true. if random long axis
c      logical*4 doranrotate !.true. if random rotations applied
      logical*4 dosmrotate ! .true. for smooth gal long rotation ! 17.04.2003

      logical*4 ISWnoise ! .true. if adding noise at ISW scale
      real*4    xISW ! factor of ISW to multiply by NSW signal
      logical*4 dovaryISW ! .true. if varying  xISW

      logical*4 doCOBEdipole       ! .true. if testing COBE dipole
      real*4    glCdip,gbCdip, xCdip,yCdip,zCdip

      logical*4 do_ivdelshift ! for slightly different iv subset e.g. BASI

      logical*4 dosiginvert ! sigma -> 1/ss2, not ss2

c     ---- 25.12.2003 ----
      logical*4 doonepair ! test one pair only 

c     ---- 02.04.2003 ----
      logical*4 doCOBE
      logical*4 doWMAP
      integer*8 j  ! to prevent overflow
      integer*8 npix8


c     ---- 09.02.2004: which type of hypotheses, e.g. T2? ----
      logical*4  doT2           !T2
      logical*4  dododec        ! Poincare dodecahedral
      logical*4  write_dodec    ! should dodec data be written?
      logical*4  dowrcircles    ! should circle data be written?
      logical*4  doonevalue     ! do only one value (e.g. for circle data)

      logical*4  ivjv_valid     ! test on validity of iv,jv


      data   glCdip,gbCdip/264.26, 48.22/ !Bennett et al. 464:L1L4, 1996

c     --- removed RE97 stuff from here ---

c     data  fun_eq/ -813.0,-446.0,-254.0,  ! e1 (x,y,z)
c     _     -490.0, 533.0, 633.0, ! e2 (x,y,z)
c     _     -153.0, 665.0, -678.0/ ! e3 (x,y,z)

      data  fun_eq/ -813.0,-490.0, -153.0, !  correct ordering
     _     -446.0, 533.0, 665.0, ! 
     _     -254.0, 633.0, -678.0/ !

      data H_0,zSLS/100.0,1100.0/

c      data flnm/'cobepix.dat'/


      indir2=indir(1:min(nindir,256))

C     BEGINNING OF STATEMENTS:

      test= .false.             !.true.              !
      cdoppwarn=.false.
      rotate=.true.             ! .false.            !

      galcut= .true.            ! .false. !
c      bcut=20.0 !30.0 !  50.0 !              ! degrees
      bcut=2.0 !20.0 !30.0 !  50.0 !              ! degrees
c     factcutnoise= 10.0 ! 1.0 ! 

      dosiginvert=.false.!.true. !

      doonepair= .false.!.true. !

      np= 2                      ! for simulations
      call chooseamps

      cmb=.true.               !.false.
      doCOBE= .false.!.true. ! 
      doWMAP= .true. !.false. !
 
      doT2=.true. !.false. !

      dododec=.true.            !.false. !
      if(dododec)doT2=.false.
      resdodec= 0.2 ! degrees  ! resolution for (l,b,and offset)
      bmindodec= 64.0          ! minimum latitude angle in degrees (90=NGP)
      bmaxdodec= 66.0           ! maximum latitude angle in degrees (90=NGP)
      glmindodec= 103.0           ! minimum longitude
      glmaxdodec= 113.0           ! maximum longitude
      thmin= 56.0
      thmax= 58.0
      angmax= 9.0! 40.0! 16.0! 14.0!  32.0 ! maximum circle angular radius in degrees
      angmin= 13.0! 1.0  !6.0 !8.0!  minimum circle angular radius in degrees
      nangle= 20! 39 ! 12! 
      twistorient= -1.0 ! \pm 1.0 for dodec, different as "control"

c     --- the "32deg" solution for zero twist ---
c      resdodec= 0.5             ! degrees  ! resolution for (l,b,and offset)
c      bmindodec= 61.0          ! minimum latitude angle in degrees (90=NGP)
c      bmaxdodec= 63.0           ! maximum latitude angle in degrees (90=NGP)
c      glmindodec= 343.0           ! minimum longitude
c      glmaxdodec= 353.0           ! maximum longitude
c      thmin= 58.0
c      thmax= 60.0
c      angmax= 40.0! 16.0! 14.0!  32.0 ! maximum circle angular radius in degrees
c      angmin= 1.0  !6.0 !8.0!  minimum circle angular radius in degrees
c      nangle= 39 ! 12! 

c     ---- "non-dodecahedral" test (ivjvlimits modified iv1=1,iv2=1) ----
      resdodec= 0.5             ! degrees  ! resolution for (l,b,and offset)
      bmindodec= 0.0          ! minimum latitude angle in degrees (90=NGP)
      bmaxdodec= 90.0           ! maximum latitude angle in degrees (90=NGP)
      glmindodec= 0.0           ! minimum longitude
      glmaxdodec= 360.0           ! maximum longitude
      thmin= 0.0
      thmax= 0.0
      angmax= 40.0! 16.0! 14.0!  32.0 ! maximum circle angular radius in degrees
      angmin= 1.0  !6.0 !8.0!  minimum circle angular radius in degrees
      nangle= 39! 2 ! 12! 
      twistorient= -1.0 ! \pm 1.0 for dodec, different as "control"


      doonevalue=.true.! .false. !
      if(doonevalue)then
         resdodec= 0.5          ! degrees  ! resolution for (l,b,and offset)
         bmindodec= 64.7        ! minimum latitude angle in degrees (90=NGP)
         bmaxdodec= 64.7        ! maximum latitude angle in degrees (90=NGP)
         glmindodec= 360.0-252.4      ! minimum longitude
         glmaxdodec= 360.0-252.4      ! maximum longitude
         angmax= 12.0!  maximum circle angular radius in degrees
         angmin= 9.0            ! minimum circle angular radius in degrees
         thmin= 57.0
         thmax= 57.0
         nangle= 10             ! 39 
         twistorient= -1.0      ! \pm 1.0 for dodec, different as "control"
      endif


c      ithsave= 12 ! interval for saving maximum values !maybe unused

      ISWnoise=.false. !.true. !
      xISW=0.0! 0.3 !0.6 !13.5/14.5 !9.0/11.0 !0.5 ! 0.2 ! 0.8 !
      ivdel=1                   !! default: ivdel=1 for 'all' circles !!
c      ivdel=6
c      ivdel=3

      nrun=2!400!200!35 !50! 1 !  2! 34 !10!  ! 34 = 360/10 * cos(18.5)

      if(dododec)then
         call getnrun(dododec, resdodec,
     _        bmindodec,bmaxdodec,glmindodec,glmaxdodec,
     _        thmin,thmax,
     _        angmin,angmax,
     _        nangle,
     _        nrun,nthdomax)
         print*,'nrun,nthdomax=',nrun,nthdomax
      endif


      q_0=0.5

c     -- 12.12.2003 boud
      w_q= -1.0

      cc_0=0.0
c     dcc=0.0025
      cc1=0.70 ! 0.0 !0.73! 0.0                   !0.617 !0.5 !0.0!
      cc2=0.8                   !622 !0.8
      ncc=1                     !                  !360 !60!
      if(ncc.gt.1)then
         dcc=(cc2-cc1)/real(ncc-1)
      else
         dcc=99.9
      endif

      ntheta= 720! 100 ! 36                !

      doRE97=.false. !.true.             !
      doRfund= .true. !.false.           ! 
      dosmrotate= .false.        ! .true. !     
      dopolerot= .false.        !.true. !
c      doranrotate=.false. !.true. 
      donudge= .false. !.true. ! 
      donud3= .false. !.true. ! 
      thnud=1.0! 0.5 ! 1.0 (great circle) degree

      dovaryISW= .false.        ! .true.          !
      xISW1=0.0
      xISW2=0.6
      doCOBEdipole=.false.!.true. !
      do_ivdelshift=.false. !.true.!

c      glong3= 53.0 !280.0 !279.0             !86.0 !14.0 !75.0 !142.0 !
c      glat3= -82.0 !37.5 ! 22.5               !-30.0 !-27.5 !18.0! 37.5 !

c     ---- for counterargument to "referee 2" ----
c      glong3= randt()*360.0
c      glat3= randt()*90.0-45.0
      print*,'glong3,glat3=',glong3,glat3
c      glong3= 280.0 !280.0 !329.0 !285.0 !100.0             ! 80.0! 
c      glat3= 37.5 !33.5 !-42.0 !-80.0 !-20.0              !50.0 !

c     WMAP
      glong3= 186.0
c      glat3= -44.0
      glat3= -41.5

c     WMAP - T3???
c      glong3= 280.5
c      glat3= -5.0

c     WMAP T3 26.12.2003 ???
c      glong3=189.3
c      glat3=48.5
c      glong3=97.5
c      glat3=1.7

c      glong3= 191.0 ! for 'bad' candidate 
c      glat3= -57.5 ! for 'bad' candidate 

c      glong3= 184.0 !264.0 
c      glat3= 8.0 !-51.0

c     -- from symT2 --
c      glong3= 329.0! 282.0 !
c      glat3=  -42.0 ! 32.0 !
c     -- after nudging --
c      glong3= 286.0
c      glat3= 35.0

c      glong3=74.5270    ! independent RE97
c      glat3= 18.46727 ! independent  RE97

c      th_d0=  328.2 !190.6 ! 0.0!   ! 328.0 for the best COBE candidate 
      th_d0=   136.0! 0.0
      th_d0=   135.5  ! 121.0! for (186.0, -41.5)
c     _     +80.0

c     WMAP T3 26.12.2003 ???
c     th_d0= 270.0 ! for (189.3,48.5) 
c      th_d0= 48.0 ! for (97.5, 1.7)


c     ---- arbitrary shifts !! test only !!----
c      th_d0=th_d0     +10.0
c      glong3= glong3+30.0 + 20.0 

c      th_d0= 40.0 ! for (280.5, -5.0)

c     --- test amr candidate from RLAGN ---
c      glong3= 281.7
c      glat3=  77.0 -90.0
c      th_d0 = 358.0

c     --- test amr candidate from RLAGN  11.07.2003---
c      glong3= 228.0
c      glat3=  75.3 -90.0
c      th_d0 = 114.2

c      th_d0= 74.1 ! for 'badsig' candidate 
c      th_d0= 42.35 
c      th_d0= 64.9 ! for (184,8) as long axis
c      th_d0= 202.0 ! approximate (not precise) value for  independent RE97

c     RE97 size = ( 1139.95    1155.63 ) for cc=0.7, om=1-cc
c     ---       cc=0  -> (540,840), cc=0.73 ->(900,1400) ---
      Rfund3_1= 9000.0 ! 18400.0! 6750.0! 8800.0 !17600.0! 9500.0! 10000.0! 540.0! 900.0 ! 1200.0 !1320.0 !1000.0 ! 600
      Rfund3_2=  19000.0!840.0! 1400.0! 1440.0 !2400.0 !1200.0 !3000.0
c      Rfund3_2= 6000.0  ! -> nrun= 181
      if(nrun.gt.1)then
         dRfund3 = (Rfund3_2 - Rfund3_1)/real(nrun-1)
      else
         dRfund3 = 9e9
      endif
      print*,'dRfund3=',dRfund3
      Rfund3=Rfund3_1

c      print*,'rSLS,drSLS=',rSLS,drSLS

      npar=1                    ! big npar not used at present


      if(doCOBEdipole)then
         xCdip= cos(gbCdip*pi/180.0) *cos(glCdip*pi/180.0)
         yCdip= cos(gbCdip*pi/180.0) *sin(glCdip*pi/180.0)
         zCdip= sin(gbCdip*pi/180.0)
      endif

      write_dodec=.false.
      if(dododec)write_dodec=.true. ! to initialise sig_th( ) etc

      do irun=1,nrun !6,6 ! ! 6,6 = test only !!
         if(irun.eq.1)then
            rotate=.false.
         else
            rotate=.true.
         endif


c     ---- read in COBE file ----

         if(irun.eq.1)then !modif 01/01/00 -> checked OK

            smooth= .true.      !.false.     !
            new=  .false.        ! .true.        !
c     -- WMAP --
            sigdeg= 1.0
            resdeg= 0.2
c     -- COBE --
            angledeg=5.0        ! 10.0 !-99.9         !
c     -- in principle any data set, though not recently debugged --
            dosmap=.false.
            dos2map=.false.
            dos2mean=.false.

            call rddata(doCOBE,doWMAP,smooth,new,sigdeg,resdeg,
     _           dosmap,dos2map,dos2mean, 
     _           indir2,nindir,
     _           galcut,bcut,
     _           Ssigma)
         endif


c     ---- SLS thickness fixed, though in principle should vary with Lambda

         drSLS= 15.0            !100.0 !
         drbig= 100.0
         drbig2= drbig*drbig

c     11.02.2004  sig being replace by par - drsigma no longer used
c         ddrsigma= 3.0*drSLS/real(nsig)
c         do i=1,nsig
c            drsigma(i)= real(i)*ddrsigma
c            drs2(i)=drsigma(i)*drsigma(i)
c         enddo
c         do icc=1,ncc
c            do i=1,mxpar
c               sigma(i,icc)=0.0
c               dmean(i,icc)=0.0
c               sfreq(i,icc)=0.0
c               corr(i,icc)=0.0
c               cdopp(i,icc)=0.0
c               cwght(i,icc)=0.0
c               if(i.eq.idlist)then
c                  do ipair=1,maxpair
c                     dlist(ipair,i,icc)=0.0
c                  enddo
c               endif
c            enddo
c         enddo


         if(dododec.and.write_dodec)then
            do ithdodec=1,mxtheta
               sig_th(ithdodec)=0.0
               dme_th(ithdodec)=0.0
               sfr_th(ithdodec)=0.0
               cor_th(ithdodec)=0.0
               cdo_th(ithdodec)=0.0
               cwg_th(ithdodec)=0.0
            enddo
            write_dodec=.false.  ! .true. first time or when writing out
         endif

         do icc=1,ncc
            cc(icc)=cc1+(real(icc-1))*dcc
            q(icc)= 0.5-1.5*cc(icc)

c     rSLS= pdWein(q_0,H_0,cc_0, zSLS)
c            rSLS= pdWein(q(icc),H_0,cc(icc), zSLS)

c     -- 12.12.2003 boud
            if(test)print*,'before dofz: q(icc),cc(icc)',q(icc),cc(icc)

            omm(icc)= 2.0*(q(icc) + cc(icc))  
c            rSLS= DEpdWeiq(omm(icc),H_0,cc(icc),w_q, zSLS)

            rSLS= real( dofz(dble(H_0), dble(omm(icc)), 
     _           dble(cc(icc)), dble(w_q), dble(zSLS)) )

            if(test)print*,'q(icc),cc(icc),rSLS=',q(icc),cc(icc),rSLS
            if(irun.eq.1.and.icc.eq.1)rSLSc=rSLS


            call hypoth(doRE97, omm(icc),H_0,cc(icc),w_q, 
     _           glong3,glat3,Rfund3,
     _           irun,nrun,
     _           dopolerot,
     _           donudge,donud3,thnud,
     _           th_d0,
     _           dovaryISW,xISW1,xISW2,
     _           ISWnoise,Ssigma,
     _           fun_eq,
     _           gl1_3,gb1_3,gl2_3,gb2_3, 
     _           xISW,
     _           bISW2)


c     ---- random rotation ----
c     - disabled by default since not really that useful - a sphere
c     with identified circles can have many equally valid choices
c     of orientations, so a genuine detection might be dismissed as
c     just one of many random events -
c            if(doranrotate)then
            if(1+1.eq.3)then
               call ranrotate(nrun,irun,fun_eq)
            endif               ! if(1+1.eq.3)then


            call hypot2(
     _           fun_eq,irun,nrun,Rfund3_1,dRfund3,rSLS,
     _           rotate,doRE97,doRfund,dosmrotate,
     _           efund,Rfund,glong,galb)


c     ---- maybe should be moved earlier if it is needed that 
c     hypoth or hypot2 
c     have an effect in the poincare dodecahedral case ----
            if(dododec)then
c               ipar=irun

               call getirun(dododec, 
     _              resdodec,
c     _              bmindodec,bmaxdodec,glmindodec,glmaxdodec,
     _              thmin,
     _              angmin,angmax,
     _              irun,
     _              glong,galb,thdodec,ang, ilb,ithdodec,iang)

               if(test)then
                  print*,'getirun: ',
     _                 glong,galb,thdodec,ang, ilb,ithdodec,iang
               endif

               gl_th(ithdodec)= glong
               gb_th(ithdodec)= galb
               thdod_th(ithdodec)= thdodec
               ang_th(ithdodec)= ang

               if(irun.eq.1)ipar=0 ! initialise ipar = 0
            else
               ithdodec= 1
            endif

            if(ithdodec.gt.mxtheta)then
               print*,'circleT2 ERROR: ithdodec, mxtheta=',
     _               ithdodec, mxtheta
               ithdodec=mxtheta
            endif


c     ---- this would probably only be needed for doing simulations ----
c            call geteinv(efund,einv) ! invert the eq fundamental matrix


c            call checkiov

c     ----- statistics ----
            if(test)print*,'starting pair list'
            

            j=0
c     sigma=0.0
c     drs2= drsigma*drsigma
            ttmean =0.0
            ttvar= 0.0

c     --- calculate the monopole, so that later it can be subtracted ---
c     --- *only* used for simulations ---
            if(.not.(cmb))then
               call monopole(cmb,doCOBE,doWMAP,
     _              ttm1,ttv1,ttsd1)
            endif

c            ttm1= 0.0001 !! test only !!

c     could set ipar to some parameter to separate diff circles sets
c     default:  ipar = 1
c            ipar=1
c            npair(ipar,icc)= 0
            npa_th(ithdodec)= 0

            itd=0

            call ivjvlimits(
     _           doT2,dododec,Rfund,rSLS,
     _           do_ivdelshift,doonepair,
     _           imax,jmax,iv1,iv2)

            mv=0
            nv=0

            do iv=iv1,iv2,ivdel ! modified 19/12/99 for BASI article
c     do iv=-imax,imax,ivdel
               do jv=0,jmax,ivdel

                  nthc(iv,jv)= 0 ! added BFR 29.01.2004

c                  if(iv.gt.0.or.jv.gt.0)then
                  call chkivjv(doT2,iv,jv, ivjv_valid)

                  if(ivjv_valid)then
                     call getvectors(
     _                    iv,jv,mv,nv,
     _                    doT2,dododec,
     _                    efund,Rfund, rSLS, ntheta,
     _                    gl_th(ithdodec),
     _                    gb_th(ithdodec),
     _                    thdod_th(ithdodec),
     _                    ang_th(ithdodec),
     _                    vx1,vy1,vz1, ax1,ay1,az1,bx1,by1,bz1,
     _                    vx2,vy2,vz2, ax2,ay2,az2,bx2,by2,bz2,
     _                    wmod, 
     _                    ivjv_valid, nth,dtheta)
                  endif

                  if(ivjv_valid)then

                     if(ipar.eq.2)then
                        dowrcircles=.true.
                     else
                        dowrcircles=.false.
                     endif

c     --- store centres (in 3d),etc. of ID'd circles ---
c                        if(irun.eq.1.and.icc.eq.1)then
                        if(dowrcircles)then
c     print*,'iv,jv=',iv,jv
                           call xyz_rda(vx1,vy1,vz1,
     _                          rr,del,alph)
                           glatc(1,iv,jv)= del
                           glongc(1,iv,jv)= alph*15.0
c                           call xyz_rda(-vx,-vy,-vz,
                           call xyz_rda(vx2,vy2,vz2,
     _                          rr,del,alph)
                           glatc(2,iv,jv)= del
                           glongc(2,iv,jv)= alph*15.0
                           wmodc(iv,jv)=  wmod
                           dthMpc(iv,jv)= dtheta*wmod
                           
                           nthc(iv,jv)= nth
                        endif
                        
                        do ith=1,nth

c                           if(test)print*,'ith, 111: ',ith
                           th= real(ith)*dtheta
                           cth= cos(th)
                           sth= sin(th)
                           
                           x1x= vx1 + wmod*(cth*ax1 + sth*bx1)
                           x1y= vy1 + wmod*(cth*ay1 + sth*by1)
                           x1z= vz1 + wmod*(cth*az1 + sth*bz1)
                           
                           if(test.and.ith.le.10)then
                              print*,'ith,x1x,x1y,x1z=',
     _                             ith,x1x,x1y,x1z
                           endif
                           
                           call xyz_rda(x1x,x1y,x1z,rr,del,alph)
                           gall1= alph*15.0
                           galb1= del
                           GC1= angle(x1x,x1y,x1z, 1.0,0.0,0.0)
c     GC1= angle(x1x,x1y,x1z, 0.0,1.0,0.0) !NOT GC

                           
                           inos=inospot(x1x,x1y,x1z)


                           if(dododec)then
c                              th=th + (1.2*pi)
                              th=th + pi + twistorient *0.2*pi
                              cth= cos(th)
                              sth= sin(th)
                           endif

                           x2x= vx2 + wmod*(cth*ax2 + sth*bx2)
                           x2y= vy2 + wmod*(cth*ay2 + sth*by2)
                           x2z= vz2 + wmod*(cth*az2 + sth*bz2)

                           if(test.and.ith.le.10)then
                              print*,'ith,x2x,x2y,x2z=',
     _                             ith,x2x,x2y,x2z
                           endif

c                           if(test)print*,'222: '
                           call xyz_rda(x2x,x2y,x2z,rr,del,alph)
                           gall2= alph*15.0
                           galb2= del
                           
c                           if(test)print*,'333: '
                           GC2= angle(x2x,x2y,x2z, 1.0,0.0,0.0)
c     GC2= angle(x2x,x2y,x2z, 0.0,1.0,0.0) !NOT GC
c                           if(test)print*,'444: '

                           
                           inos=inos.or.inospot(x2x,x2y,x2z)
                              
c                           if(test)print*,'555: '

                           if(.not.(galcut).or.
     _                          (abs(galb1).gt.bcut
     _                          .and.abs(galb2).gt.bcut 
c     _                             .and.galb1.gt.0.0 ! SG hemisph EXcluded
c     _                             .and.galb2.gt.0.0 ! SG hemisph EXcluded
c     _                     .and.(GC1.gt.pi/3.0.or.galb1.gt.0.0) !+90 EXcluded
c     _                     .and.(GC2.gt.pi/3.0.or.galb2.gt.0.0) !+90 EXcluded
     _                          .and.GC1.gt.pi/9.0 ! GC EXcluded
     _                          .and.GC2.gt.pi/9.0 ! GC EXcluded
c     _                          .and.(GC1.lt.pi/3.0  ! GC included
c     _                              .or.GC2.lt.pi/3.0) ! GC included
c     _                          .and.(.not.(inos)) ! hot spots EXcluded
c     _                          .and.((inos)) ! hot spots included
c     _                          .and.(.false.)  ! test only !
c     _                             .and.(.not.(isMage(gall1,galb1))) !Mag-
c     _                             .and.(.not.(isMage(gall2,galb2))) ! stream
c     _                             .and.(.not.(antiGC(gall1,galb1))) !antiGC
c     _                             .and.(.not.(antiGC(gall2,galb2))) ! 
c     _                             .and.(.not.(SGcut(gall1,galb1))) !arbitr.
c     _                             .and.(.not.(SGcut(gall2,galb2))) ! SG cut
     _                          )
     _                          ) then


c     if(1.lt.iv.and.iv.lt.5.and.jv.eq.0)then
c     print*,'TEST out'
c     print*,'x1x,x1y,x1z=',x1x,x1y,x1z
c     print*,'x2x,x2y,x2z=',x2x,x2y,x2z
c     endif
c     !! test only !!
c     gall1=gall1 + 23.0
c     gall2=gall2 + 58.0
c     galb1=galb1 +38.0
c     galb2=galb2 -67.0

                              noise=.false.
                              if(cmb)then
                                 if(doCOBE)then
                                    tt1= cobeDMR(gall1,galb1)
                                    tt2= cobeDMR(gall2,galb2)
                                 elseif(doWMAP)then
                                    tt1= wmaplin(gall1,galb1)
                                    tt2= wmaplin(gall2,galb2)
                                 endif
                              else
                                 tt1= delTgal(gall1,galb1)
                                 tt2= delTgal(gall2,galb2)
                                 
c     -- following two lines needed to remove monopole 
c     should affect  corr  and  cdopp  but not  dmean nor sigma --
                                 tt1=tt1-ttm1
                                 tt2=tt2-ttm1
                                 
                              endif


                              noise=.true.
                              if(doCOBE)then
                                 s1= cobeDMR(gall1,galb1)
                                 s2= cobeDMR(gall2,galb2)
                              elseif(doWMAP)then
                                 s1= wmaplin(gall1,galb1)
                                 s2= wmaplin(gall2,galb2)
                              endif

c     ---- add ISW noise in quadrature 20/08/99 ----
                              if(ISWnoise)then
                                 s1= sqrt(bISW2+ s1*s1)
                                 s2= sqrt(bISW2+ s2*s2)
                              endif

c                              if(irun.eq.1.and.icc.eq.1)then
                              if(dowrcircles)then
                                 tt1c(ith,iv,jv)= tt1
                                 tt2c(ith,iv,jv)= tt2
                                 s1c(ith,iv,jv)= s1
                                 s2c(ith,iv,jv)= s2
                                 glpix1(ith,iv,jv)= gall1
                                 gbpix1(ith,iv,jv)= galb1
                                 glpix2(ith,iv,jv)= gall2
                                 gbpix2(ith,iv,jv)= galb2
                              endif
c     s1=1.0
c     s2=1.0

                              xx= tt1-tt2
                              yy= max(s1*s1+s2*s2,1e-20) !
                              if(test.and.yy.lt.2e-20)print*,
     _                             'WARNING s1,s2=',s1,s2 
                              ss2= xx*xx /yy 
                           
                              if(dosiginvert)then
                                 if(ss2.ge.1e-4)then ! hard limit
                                    ss2=1.0/ss2
                                 else
                                    ss2=0.0
                                 endif
                              endif
                           
c     wt= 1.0/sqrt(yy)
                              ss= xx /sqrt(yy)
                           
                              if(.not.(doCOBEdipole))then
                                 dd= dot(x1x,x1y,x1z, 
     _                                x2x,x2y,x2z)
     _                                /(rSLS*rSLS)
                              else
                                 dd= dot( (x1x)/rSLS, 
     _                                (x1y)/rSLS, 
     _                                (x1z)/rSLS, 
     _                                xCdip,yCdip,zCdip) *
     _                                dot( (x2x)/rSLS, 
     _                                (x2y)/rSLS, 
     _                                (x2z)/rSLS, 
     _                                xCdip,yCdip,zCdip)
                              endif
                              

c     if(1.lt.iv.and.iv.lt.5.and.jv.eq.0)then
c     print*,'TEST out'
c     print*,'x1x,x1y,x1z=',x1x,x1y,x1z
c     print*,'x2x,x2y,x2z=',x2x,x2y,x2z
c     print*,'dd,rSLS*rSLS=',dd,rSLS*rSLS
c     endif

c                              sigma(ipar,icc)= sigma(ipar,icc)+ ss2
c                              dmean(ipar,icc)= dmean(ipar,icc)+ ss
c                              sfreq(ipar,icc)= sfreq(ipar,icc) +1.0

                              sig_th(ithdodec)= sig_th(ithdodec)+ ss2
                              dme_th(ithdodec)= dme_th(ithdodec)+ ss
                              sfr_th(ithdodec)= sfr_th(ithdodec) +1.0
                              
                              if(abs(dd).gt.1.0)print*,'WARNING,',
     _                             'dd = ',dd,' > 1.0 '
                              
c                              corr(ipar,icc)=corr(ipar,icc)
cc     _                             + wt*wt*tt1*tt2
c     _                             + 2.0* tt1*tt2 ! CSS98b
c                              cdopp(ipar,icc)=cdopp(ipar,icc) 
cc     _                             + wt*wt*tt1*tt2 *dd
c     _                             + 2.0*tt1*tt2 *dd !*dd !  cf. CSS98b
cc     cwght(ipar,icc)=cwght(ipar,icc)+ yy
cc     CSS98b version: 
c                              cwght(ipar,icc)=cwght(ipar,icc)+ 
c     _                             tt1*tt1+tt2*tt2



                              cor_th(ithdodec)=cor_th(ithdodec)
c     _                             + wt*wt*tt1*tt2
     _                             + 2.0* tt1*tt2 ! CSS98b
                              cdo_th(ithdodec)=cdo_th(ithdodec) 
c     _                             + wt*wt*tt1*tt2 *dd
     _                             + 2.0*tt1*tt2 *dd !*dd !  cf. CSS98b
c     cwg_th(ithdodec)=cwg_th(ithdodec)+ yy
c     CSS98b version: 
                              cwg_th(ithdodec)=cwg_th(ithdodec)+ 
     _                             tt1*tt1+tt2*tt2



c                              if(test)print*,'666: '

c     ---- pair list for KS test ----
c                              npair(ipar,icc)=npair(ipar,icc)+1
                              npa_th(ithdodec)=npa_th(ithdodec)+1
                              itd=itd+1
c                              dlist(npair(ipar,icc),idlist,icc)= ss
                              dlist(npa_th(ithdodec),idlist,icc)= ss
                           
c                              if(cdoppwarn.and.(abs(cdopp(ipar,icc))
c     _                             .gt.abs(corr(ipar,icc))))
c     _                             print*,'WARNING,',
c     _                             'cdopp(ipar,icc),corr(ipar,icc)='
c     _                             ,cdopp(ipar,icc),corr(ipar,icc)
                              if(cdoppwarn.and.(abs(cdo_th(ithdodec))
     _                             .gt.abs(cor_th(ithdodec))))
     _                             print*,'WARNING,',
     _                             'cdo_th(ithdodec),cor_th(ithdodec)='
     _                             ,cdo_th(ithdodec),cor_th(ithdodec)
                           
                           
                              j=j+1
                              ttmean=ttmean+tt1+tt2
                              ttvar=ttvar + tt1*tt1 +tt2*tt2
                              
                           else
c                              if(irun.eq.1.and.icc.eq.1)then
                              if(dowrcircles)then
                                 tt1c(ith,iv,jv)= -9.99
                                 tt2c(ith,iv,jv)= -9.99
                                 s1c(ith,iv,jv)= -9.99
                                 s2c(ith,iv,jv)= -9.99
                              endif
                           endif !  if( (.not.(galcut).or. ...
                           
c                           if(test)print*,'777: '
                        enddo   ! do ith=1,nth
                        
c                        print*,'iv,jv,nth,npair(ipar,icc)=',
c     _                       iv,jv,nth,npair(ipar,icc)
                        if(test)
     _                       print*,'iv,jv,nth,npa_th(ithdodec)=',
     _                       iv,jv,nth,npa_th(ithdodec)
                        
                     
ccc   endif      ! if(nth.ge.1)then
                     
                  endif      ! if(ivjv_valid)then

cc                     endif      !                     if(vmod.lt.rSLS/2.0)then
cc                  endif         ! if(iv.gt.0.or.jv.gt.0)then

               enddo         ! do jv=0,jmax
            enddo               ! do iv=-imax,imax

c            print*,':: irun,nth,npair(ipar,icc)=',
c     _           irun,nth,npair(ipar,icc)

            if(10000*((irun-1)/10000).eq.irun)then
               print*,':: irun,nth,npa_th(ithdodec)=',
     _              irun,nth,npa_th(ithdodec)
            endif

            ttmean=ttmean/2.0/real(j)
            
            ttvar= ttvar/2.0/real(j-1) - ttmean*ttmean 
            ttsd=-99.9
            if(ttvar.gt.0.0)ttsd=sqrt(ttvar)

            if(test)then
               print*,'j=',j,' npix*(npix+1)/2=',npix*(npix-1)/2
               npix8=npix
               print*,'j=',j,' npix8*(npix8+1)/2=',npix8*(npix8-1)/2
               print*,'zero point stats: ttmean,ttvar,ttsd=',
     _              ttmean,ttvar,ttsd

               print*,'unnormalised: dmean(.,icc)=',
     _              (dme_th(ithdodec))
               print*,'unnormalised: sigma(.,icc)=',
     _              (sig_th(ithdodec))
               print*,'unnormalised: corr(.,icc)=',
     _              (cor_th(ithdodec))
               print*,'unnormalised: cdopp(.,icc)=',
     _              (cdo_th(ithdodec))
               print*,'sfreq=',(sfr_th(ithdodec))
               print*,'cwght=',(cwg_th(ithdodec))
               print*,' '
            endif

c            do ipar=1,npar

c            if(sfreq(ipar,icc).gt.0.001)then
c               dmean(ipar,icc)=dmean(ipar,icc)/sfreq(ipar,icc)
c               corr(ipar,icc)=corr(ipar,icc)/cwght(ipar,icc)
c               cdopp(ipar,icc)=cdopp(ipar,icc)/cwght(ipar,icc)
c               
c               sigma(ipar,icc)=sigma(ipar,icc)/sfreq(ipar,icc)
c            endif
            if(sfr_th(ithdodec).gt.0.001)then
               dme_th(ithdodec)=dme_th(ithdodec)/sfr_th(ithdodec)
               cor_th(ithdodec)=cor_th(ithdodec)/cwg_th(ithdodec)
               cdo_th(ithdodec)=cdo_th(ithdodec)/cwg_th(ithdodec)
               
               sig_th(ithdodec)=sig_th(ithdodec)/sfr_th(ithdodec)
            endif

c            tol=1e-4
c            if(abs(sigma(ipar,icc)-sig_th(ithdodec)).gt.tol)then
c               print*,'## ',irun,sigma(ipar,icc),sig_th(ithdodec)
c            endif
            

c     if(ttsd.gt.0.0)then
c     cdopp(ipar,icc)=cdopp(ipar,icc)/ttmean/ttmean -1.0
c     corr(ipar,icc)=corr(ipar,icc)/ttmean/ttmean -1.0
c     endif

c            enddo

            write_dodec= dododec.and.(ithdodec.eq.nthdomax)

            if(test.and.irun.le.8)
     _           print*,'A-- sigma::',
     _           irun,ithdodec,sig_th(1),sig_th(2)

c            if(2+2.eq.5.and.(.not.(dododec)))then
            if(.not.(dododec))then
               ipar=irun
               dmean(ipar,icc)=dme_th(ithdodec)
               sigma(ipar,icc)=sig_th(ithdodec)
               corr(ipar,icc)= cor_th(ithdodec)
               cdopp(ipar,icc)=cdo_th(ithdodec)

               sfreq(ipar,icc)=sfr_th(ithdodec)
               cwght(ipar,icc)=cwg_th(ithdodec)

               npair(ipar,icc)=npa_th(ithdodec)
            elseif(write_dodec)then
               ithcmax=1 ! maximise corr
               ithsmin=1 ! minimise sigma
               cmax= -9.9e9
               smin= +9.9e9
               do ith=1,nthdomax
                  if(cor_th(ith).gt.cmax)then
                     ithcmax=ith
                     cmax=cor_th(ith)
                  endif
                  if(sig_th(ith).lt.smin)then
                     ithsmin=ith
                     smin=sig_th(ith)
                  endif
               enddo
               if(test.and.irun.le.8)then
                  print*,'write_dodec: ipar,ithcmax,ithsmin',
     _                 ipar,ithcmax,ithsmin
               endif
c     ---- sigma min ----
               ipar=ipar+1

               if(test.and.irun.le.8)then
                  print*,'A- sigma::',sig_th(1),sig_th(2)
                  print*,'AA sigma::',
     _                 ipar,icc,ithsmin,sig_th(ithsmin),
     _                 sigma(ipar,icc)
               endif

               dmean(ipar,icc)=dme_th(ithsmin)
               sigma(ipar,icc)=sig_th(ithsmin)
               corr(ipar,icc)= cor_th(ithsmin)
               cdopp(ipar,icc)=cdo_th(ithsmin)

               sfreq(ipar,icc)=sfr_th(ithsmin)
               cwght(ipar,icc)=cwg_th(ithsmin)

               npair(ipar,icc)=npa_th(ithsmin)

               glpar(ipar)= gl_th(ithsmin)
               gbpar(ipar)= gb_th(ithsmin)
               thpar(ipar)= thdod_th(ithsmin)
               angpar(ipar)= ang_th(ithsmin)

               if(test.and.irun.le.8)then
                  print*,'AAA sigma::',
     _                 ipar,icc,ithsmin,sig_th(ithsmin),
     _                 sigma(ipar,icc)
               endif
               if(test.and.irun.le.8)then
                  print*,'A sigma: ',ipar,
     _                 ((sigma(j,k),j=1,6),k=1,1)
               endif

c     ---- corr max ----
               ipar=ipar+1
               dmean(ipar,icc)=dme_th(ithcmax)
               sigma(ipar,icc)=sig_th(ithcmax)
               corr(ipar,icc)= cor_th(ithcmax)
               cdopp(ipar,icc)=cdo_th(ithcmax)

               sfreq(ipar,icc)=sfr_th(ithcmax)
               cwght(ipar,icc)=cwg_th(ithcmax)

               npair(ipar,icc)=npa_th(ithcmax)

               glpar(ipar)= gl_th(ithcmax)
               gbpar(ipar)= gb_th(ithcmax)
               thpar(ipar)= thdod_th(ithcmax)
               angpar(ipar)= ang_th(ithcmax)
               if(test.and.irun.le.8)print*,'ipar = ',ipar
               if(test.and.irun.le.8)then
                  print*,'B sigma: ',ipar,
     _                 ((sigma(j,k),j=1,6),k=1,1)
               endif
            endif



            if(sfrbig(icc).gt.1.001)then
c     sigbig(icc)= sqrt( sigbig(icc)/(sfrbig(icc)-1.0) )
               sigbig(icc)= sqrt( sigbig(icc)/(sfrbig(icc)-0.0) )
            endif
            if(test)print*,' '
            if(test)print*,'icc,sigbig(icc)=',icc,sigbig(icc)

            if(test)then
               print*,'NORMALISED: dmean(.,icc)='
     _              ,(dme_th(ithdodec))
               print*,'NORMALISED: sigma(.,icc)='
     _              ,(sig_th(ithdodec))
               print*,'NORMALISED: corr(.,icc)='
     _              ,(cor_th(ithdodec))
               print*,'NORMALISED: cdopp(.,icc)='
     _              ,(cdo_th(ithdodec))

            endif

         enddo                  !             do icc=1,ncc

         if(ipar.eq.1)then
            ch3='_RE'
         else
            write(ch3,'(i3)')ipar-1
            if(ch3(1:1).eq.' ')ch3(1:1)='_'
            if(ch3(2:2).eq.' ')ch3(2:2)='_'
         endif


         if(ipar.le.mxwrite.and.
     _        ( (.not.(dododec)).or. write_dodec)
     _        )then
c     ---- write sigma's to file ----
            open(1,file='sigma'//ch3//'.dat',access='sequential',
     _           form='unformatted',status='unknown')
            write(1)glong,galb,thetadeg,Rfund(1),Rfund(2),Rfund(3),
     _           gl1_3,gb1_3,gl2_3,gb2_3,xISW,itd
c     --- WARNING: potential conflict in size of  drsigma   with plotsig.f --
c         write(1)ncc,cc1,cc2,H_0,drSLS,npar,drsigma,maxpair
c         write(1)dmean
c         write(1)sigma
c         write(1)sfreq
c         write(1)corr
c         write(1)cdopp
c         write(1)cwght
c         write(1)npair

            ipar1=ipar
            ipar2=ipar
            write(1)ncc,cc1,cc2,H_0,drSLS,ipar1,ipar2,maxpair
            
            write(1)((dmean(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            write(1)((sigma(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            write(1)((sfreq(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            write(1)(( corr(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            write(1)((cdopp(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            write(1)((cwght(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            write(1)((npair(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
            
            write(1)(((dlist(ip,i,icc),ip=1,maxpair),
     _           i=1,mxdlist),icc=1,ncc)
            close(1)


               if(test.and.irun.le.8)then
                  print*,'C sigma: ',ipar,
     _                 ((sigma(j,k),j=1,6),k=1,1)
               endif

         endif

         if(test)then
            print*,'TEST B nthc(1,0),glatc(1,1,0),glongc(1,1,0)='
            print*,nthc(1,0),glatc(1,1,0),glongc(1,1,0)
         endif

c     ---- write circle data to file ----
c         if(ipar.le.2.and.ipar.le.mxwrite.and.
c     _        ( (.not.(dododec)).or. write_dodec)
c     _        )then
c         if(irun.eq.1)then
         if(dowrcircles)then
            nparcircles= 1

            open(1,file='circleinfo'//ch3//'.dat',
     _           access='sequential',
     _           form='unformatted',status='unknown')
c            write(1)ncc,cc1,cc2,H_0,drSLS,npar,drsigma
c            write(1)ncc,cc1,cc2,H_0,drSLS,nparcircles,drsigma !24.02.2004
            write(1)ncc,cc1,cc2,H_0,drSLS,nparcircles 
            write(1)iimax,jjmax,maxtheta, imax,jmax
            write(1)nthc,glatc,glongc,wmodc,rSLSc,dthMpc
            write(1)tt1c,tt2c,s1c,s2c
c     ---- 24.02.2004 ----
            write(1)glpix1,gbpix1,glpix2,gbpix2
            close(1)
         endif
      enddo            !      do irun=1,nrun


      npar=ipar

c     ---- final file ----
      ch3='all'
      open(1,file='sigma'//ch3//'.dat',access='sequential',
     _     form='unformatted',status='unknown')
      write(1)glong,galb,thetadeg,Rfund(1),Rfund(2),Rfund(3),
     _     gl1_3,gb1_3,gl2_3,gb2_3,xISW,itd

      ipar1=1
      ipar2=npar
      write(1)ncc,cc1,cc2,H_0,drSLS,ipar1,ipar2,maxpair
            
      write(1)((dmean(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      write(1)((sigma(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      write(1)((sfreq(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      write(1)(( corr(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      write(1)((cdopp(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      write(1)((cwght(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      write(1)((npair(i,icc),i=ipar1,ipar2),icc=1,maxcurv)

      write(1)(glpar(i),i=ipar1,ipar2)
      write(1)(gbpar(i),i=ipar1,ipar2)
      write(1)(thpar(i),i=ipar1,ipar2)
      write(1)(angpar(i),i=ipar1,ipar2)

      if(test)then
         print*,'dme',((dmean(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
         print*,'sig',((sigma(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
         print*,'sfr',((sfreq(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
         print*,'cor',(( corr(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
         print*,'cdo',((cdopp(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
         print*,'cwg',((cwght(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
         print*,'npa',((npair(i,icc),i=ipar1,ipar2),icc=1,maxcurv)
      endif
      
      write(1)(((dlist(ip,i,icc),ip=1,maxpair),
     _     i=1,mxdlist),icc=1,ncc)
      close(1)



      end

c     !! dot( ) moved to rdCOBE.f
c     !! dotv, crossv, orth_axes to orthaxes.f 


